/* simserv : works with the gamecube simulator
 * Copyright (C) 2004  Alexandre Boeglin <alex@boeglin.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <zlib.h>

#define GCSIM_PORT 8191
#define BUFFER_SIZE 1028

typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;
typedef unsigned long int u64;

// compressed_file_read reads from fd_in and fills file_buffer
// fd_in is the compressed input file
// header is the input file header
// file_buffer is the output buffer
// block_size is the block_size
// stream_offset is the offset in fd_in
// stream_size is the size to read
int compressed_file_read(int fd_in, u32 *header, u8 *file_buffer, u32 block_size, u32 stream_offset, u32 stream_size){

  int i, error_test;

  u32 stream_end, first_block, last_block, actual_read_size;
  u64 actual_uncompressed_size;
  u8 *in_buffer, *out_buffer, *concat_buffer;
  
  // prepare uncompression
  stream_end = stream_offset + stream_size - 1;
  first_block = stream_offset / block_size;
  last_block = stream_end / block_size;

  error_test = lseek(fd_in, header[2 * first_block], SEEK_SET);
  if(error_test == -1)
    return 1;

  // allocate buffer
  in_buffer = malloc(block_size * 2);
  if(in_buffer == NULL)
    return 2;
  out_buffer = malloc(block_size);
  if(out_buffer == NULL)
    return 3;
  concat_buffer = malloc(block_size * (last_block - first_block + 1));
  if(concat_buffer == NULL)
    return 4;

  // uncompression loop
  for(i = first_block; i <=last_block; i++){
    // read data
    actual_read_size = read(fd_in, in_buffer, header[2 * i + 1]);
    if(actual_read_size == -1)
      return 5;

    // uncompress data
    actual_uncompressed_size = block_size;
    error_test = uncompress(out_buffer, &actual_uncompressed_size, in_buffer, actual_read_size);
    if(error_test != Z_OK){
      return 6;
    }

    // write data
    memcpy(concat_buffer + block_size * (i - first_block), out_buffer, actual_uncompressed_size);
  }
  // extract query part from buffer
  memcpy(file_buffer, concat_buffer + stream_offset % block_size, stream_size);

  // free buffer
  free(in_buffer);
  free(out_buffer);
  free(concat_buffer);

  return 0;
}

// uncompressed_file_read reads from fd_in and fills file_buffer
// in this case, header and block_size are ignored. for other parameters, please have a look at compressed_file_read
int uncompressed_file_read(int fd_in, u32 *header, u8 *file_buffer, u32 block_size, u32 stream_offset, u32 stream_size){
  
  int error_test;
  
  // seek to offset
  error_test = lseek(fd_in, stream_offset, SEEK_SET);
  if(error_test == -1)
    return 1;

  // read data
  error_test = read(fd_in, file_buffer, stream_size);
  if(error_test == -1)
    return 2;

  return 0;
}

int main(int argc, char **argv){

  char c = '.', ident_string[9] = "ngc-comp", comp_string[9] = "12345678";
  int fd_in, my_socket, i, addr_size, error_test;
  struct sockaddr_in my_addr, gc_addr;

  int (*file_read)();

  u16 packets_left, current_batch, old_batch;
  u32 stream_offset, stream_size, block_size, block_number, header_size, actual_send_size, server_response, batch_size, current_batch_size;
  u8 net_buffer[BUFFER_SIZE], *file_buffer;
  u32 *header;

  // check args
  if(argc != 2)
    exit(1);

  // open file
  fd_in = open(argv[1], O_RDONLY);
  if(fd_in == -1)
    exit(2);

  // check ident string
  error_test = read(fd_in, comp_string, 8);
  if(error_test == -1)
    exit(3);
  if(memcmp(ident_string, comp_string, 8) != 0){ // strings differ, this is not a compressed file
    // set variables
    file_read = uncompressed_file_read;
    header = NULL;
  }
  else { // string match, this is a compressed file
    // set variables
    file_read = compressed_file_read;

    // get block size and block number
    error_test = read(fd_in, &block_size, sizeof(block_size));
    if(error_test == -1)
      exit(4);
    block_size = ntohl(block_size);
    error_test = read(fd_in, &block_number, sizeof(block_number));
    if(error_test == -1)
      exit(5);
    block_number = ntohl(block_number);

    // get header
    header_size = sizeof(u32) * (2 * block_number);
    header = malloc(header_size);
    if(header == NULL)
      exit(6);
    error_test = read(fd_in, header, header_size);
    if(error_test == -1)
      exit(7);
    for(i = 0; i < (2 * block_number); i++)
      header[i] = ntohl(header[i]);
  }

  // prepare buffer
  batch_size = (BUFFER_SIZE - 4) * (sizeof(u16) * 8); // 1024 * 16
  file_buffer = malloc(batch_size);
  if(file_buffer == NULL)
    exit(8);

  // prepare network
  memset(&my_addr, 0, sizeof(my_addr));
  my_addr.sin_family = AF_INET;
  my_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  my_addr.sin_port = htons(GCSIM_PORT);
  addr_size = sizeof(my_addr);

  // bind socket
  my_socket = socket(AF_INET, SOCK_DGRAM, 0);
  if(my_socket == -1)
    exit(9);
  error_test = bind(my_socket, (struct sockaddr *)&my_addr, addr_size);
  if(error_test == -1)
    exit(10);

  // stream loop
  while(1){
    // initialize variables
    old_batch = 0xFFFF; // some impossible value, so that first batch of a request will always be marked as new
    current_batch_size = batch_size;

    // wait for query (0xF80000, offset, size)
    memset(&gc_addr, 0, sizeof(gc_addr));
    error_test = recvfrom(my_socket, net_buffer, BUFFER_SIZE, 0, (struct sockaddr *)&gc_addr, &addr_size);
    if(error_test == -1)
      exit(11);

    // get query infos
    stream_offset = ((u32 *)net_buffer)[1];
    stream_offset = ntohl(stream_offset);
    stream_size = ((u32 *)net_buffer)[2];
    stream_size = ntohl(stream_size);

    // send acknowledgement (0x00000001)
    server_response = 1;
    server_response = htonl(server_response);
    error_test = sendto(my_socket, &server_response, 4, 0, (struct sockaddr *)&gc_addr, addr_size);
    if(error_test == -1)
      exit(12);

    // process packet batches
    while(1){
      // get query (packets_left, current_batch)
      error_test = recvfrom(my_socket, net_buffer, BUFFER_SIZE, 0, (struct sockaddr *)&gc_addr, &addr_size);
      if(error_test == -1)
        exit(13);

      // get query infos
      packets_left = ((u16 *)net_buffer)[0];
      packets_left = ntohs(packets_left);
      current_batch = ((u16 *)net_buffer)[1];
      current_batch = ntohs(current_batch);

      if(packets_left == 0 && current_batch == 0) // no more to send
        break;

      // read data if it's a new batch
      if(current_batch != old_batch){
        current_batch_size = stream_size - batch_size * current_batch;
        if(current_batch_size > batch_size)
          current_batch_size = batch_size;
        error_test = file_read(fd_in, header, file_buffer, block_size, stream_offset + batch_size * current_batch, current_batch_size);
        if(error_test != 0)
          exit(100 + error_test);
      }

      // send data
      for(i = 0; i < 16; i++){
        if(packets_left & (1<<i)){ // packet i is requested
          // prepare packet header
          net_buffer[0] = (1<<i) / 256;
          net_buffer[1] = (1<<i) % 256;
          net_buffer[2] = current_batch / 256;
          net_buffer[3] = current_batch % 256;

          // send data (packet_num, batch, data)
          actual_send_size = current_batch_size - (BUFFER_SIZE - 4) * i;
          if(actual_send_size > 1024)
            actual_send_size = 1024;
          memcpy(net_buffer + 4, file_buffer + (BUFFER_SIZE - 4) * i, actual_send_size);
          error_test = sendto(my_socket, net_buffer, actual_send_size + 4, 0, (struct sockaddr *)&gc_addr, addr_size);
          if(error_test == -1)
            exit(14);
        }
      }
      // remember this batch (usefull when packets are lost)
      old_batch = current_batch;
    }
    // warn user
    write(1, &c, 1);
  }
  return 0;
}
